import unicodecsv as csv
import json
import abc
import terminaltables
import dpath.util as dpath
from textwrap import wrap


def for_format(format, file, **kwargs):
    json_opts = kwargs.pop('json', None)
    if format == 'json':
        json_opts = json_opts or {
            "indent": 4
        }
        return JSONWriter(file, json=json_opts, **kwargs)
    elif format == 'csv':
        json_opts = json_opts or {
            "indent": None
        }
        return CSVWriter(file, json=json_opts, **kwargs)
    elif format == 'tty':
        json_opts = json_opts or {
            "indent": 0
        }
        return TerminalWriter(file, json = json_opts, **kwargs)

class BaseWriter:
    __metaclass__ = abc.ABCMeta

    def __init__(self, file, **kwargs):
        self.file = file
        self.mapping = kwargs.pop('mapping', None)
        self.json_opts = kwargs.pop('json', None)

    def start(self):
        return

    @abc.abstractmethod
    def write_entries(self, entries):
        return

    def finish(self):
        return

    def lookup(self, o, path, **kwargs):
        try:
            if hasattr(path, '__call__'):
                result = path(o)
            else:
                result = dpath.get(o, path)
        except KeyError:
            return ""
        json_opts = kwargs.pop('json', None)
        if isinstance(result, dict) and json_opts:
            result = json.dumps(result, **json_opts)
        else:
            result = result
        return result


class CSVWriter(BaseWriter):
    def __init__(self, file, **kwargs):
        super(CSVWriter, self).__init__(file, **kwargs)
        self.writer = csv.DictWriter(file, self.mapping.keys())

    def start(self):
        self.writer.writeheader()

    def write_entries(self, entries):
        if entries is None: return
        for entry in entries:
            row = { key: self.lookup(entry, path, json=self.json_opts)
                    for key, path in self.mapping.iteritems() }
            self.writer.writerow(row)


class JSONWriter(BaseWriter):
    def __init__(self, file, **kwargs):
        super(JSONWriter, self).__init__(file, **kwargs)

    def start(self):
        self.file.write('[\n')
        self.is_first_entries = True

    def write_entries(self, entries):
        if entries is None: return
        for entry in entries:
            if self.is_first_entries:
                self.is_first_entries = False
            else:
                self.file.write(',')

            if self.mapping:
                entry = {key: self.lookup(entry, path) for key, path in self.mapping.iteritems() }

            json.dump(entry, self.file, **self.json_opts)

    def finish(self):
        self.file.write('\n]\n')


class TerminalWriter(BaseWriter):
    def __init__(self, file, **kwargs):
        super(TerminalWriter, self).__init__(file, **kwargs)
        self.data = [[key.upper() for key in self.mapping.keys()]]
        self.json_opts = kwargs.pop('json', None)
        self.max_width = kwargs.pop('max_width', None)

    def start(self):
        pass

    def write_entries(self, entries):
        def safe_unicode(o, *args):
            try:
                return unicode(o, *args)
            except UnicodeDecodeError:
                return unicode(str(o).encode('string_escape'))
        def map_entry(o):
            return [safe_unicode(self.lookup(o, path, json=self.json_opts))
                    for _,path in self.mapping.items()]

        if entries is None: return
        if self.mapping:
            self.data.extend([map_entry(o) for o in entries])
        else:
            self.data.extend(entries)

    def finish(self):
        table = terminaltables.AsciiTable(self.data)
        # table.inner_column_border=False
        # table.inner_heading_row_border=False
        # table.outer_border=False
        last_column = len(self.mapping) - 1
        max_width = self.max_width or table.column_max_width(last_column)
        for entry in table.table_data:
            entry[last_column] = '\n'.join(wrap(entry[last_column], max_width))
        self.file.write(table.table)
        self.file.write('\n')



import dpath.util as dpath
import requests
import click
import helium
import util
import timeseries as ts

pass_service=click.make_pass_decorator(helium.Service)

@click.group()
def cli():
    """Operations on cloud-scripts.
    """
    pass

def _tabulate(result):
    util.tabulate(result, [
        ('id', util.shorten_json_id),
        ('state', 'attributes/state'),
        ('name', 'attributes/name'),
        ('files', util.map_script_filenames),
        ('error', 'attributes/error/error')
    ])


@cli.command()
@pass_service
def list(service):
    """List all known cloud-scripts.

    Lists all cloud-scripts for the organization.
    """
    _tabulate(service.get_cloud_scripts().get('data'))

@cli.command()
@click.argument('script')
@click.option('--name',
              help="the new name for the script")
@pass_service
def start(service, script, **kwargs):
    """Starts a given cloud-script.

    Starts the SCRIPT with the given id.
    """
    script = util.lookup_resource_id(service.get_cloud_scripts, script)
    _tabulate([service.update_cloud_script(script, state="running", **kwargs).get('data')])

@cli.command()
@click.argument('script')
@click.option('--name',
              help="the new name for the script")
@pass_service
def stop(service, script, **kwargs):
    """Stop a given cloud-script.

    Stops the SCRIPT with the given id.
    """
    script = util.lookup_resource_id(service.get_cloud_scripts, script)
    _tabulate([service.update_cloud_script(script, state="stopped", **kwargs).get('data')])


@cli.command()
@click.argument('script')
@pass_service
def delete(service, script):
    """Delete a given cloud-script.

    Deletes the cloud-SCRIPT with the given id.
    """
    script = util.lookup_resource_id(service.get_cloud_scripts, script)
    result = service.delete_cloud_script(script)
    click.echo("Deleted: " + script.encode('ascii')
               if result.status_code == 204 else result)


@cli.command()
@click.argument('script')
@ts.options()
@pass_service
def timeseries(service, script, **kwargs):
    """List readings for a cloud-script.

    Lists one page of readings for a given SCRIPT.
    """
    script = util.lookup_resource_id(service.get_cloud_scripts, script)
    data = service.get_cloud_script_timeseries(script, **kwargs).get('data')
    ts.tabulate(data, **kwargs)


@cli.command()
@click.argument('script')
@click.argument('file')
@pass_service
def show(service, script, file):
    """Gets a script file from a given cloud-script.

    Fetches a FILE from a given cloud-SCRIPT.
    """
    script = util.lookup_resource_id(service.get_cloud_scripts, script)
    json = service.get_cloud_script(script).get('data')
    file_urls = [f.encode('utf-8') for f in dpath.get(json, 'meta/scripts')]
    names = dict(zip(util.extract_script_filenames(file_urls), file_urls))
    file_url = names[file]
    click.echo(requests.get(file_url).text)


@cli.command()
@click.argument('file', nargs=-1)
@click.option('--main', type=click.Path(exists=True),
              help="The main file for the script")
@click.option('--name',
              help="the name for the script")
@click.option('--state', type=click.Choice(['running', 'stopped']),
              default='running',
              help="the starting state for the script (default 'running')")
@pass_service
def deploy(service, file, **kwargs):
    """Deploy  a cloud-script.

    Submits a deploy request of one ore more FILEs.
    Optionally specify a name and the starting state of the script once
    it is deployed.

    If the --main option is specified the given file is used as the `user.lua`,
    i.e. the main user script for the deployment. The file may  be part of
    the list of files given to make it easier to specify wildcards.

    Note: One of the given files _must_ be called user.lua if the --main option is not given.
    This file will be considered the primary script for the deploy.
    """
    deploy=service.deploy_cloud_script(file, **kwargs).get('data')
    _tabulate([deploy])


@cli.command()
@click.argument('script')
@pass_service
def status(service, script):
    """Displays current status information for a script.

    Display status information a given SCRIPT. If the script is in an error
    condition the error details are displayed.
    """
    script = util.lookup_resource_id(service.get_cloud_scripts, script)
    data = service.get_cloud_script(script).get('data')
    click.echo('Status: ' + dpath.get(data, "attributes/state"))
    try:
        error = dpath.get(data, "attributes/error")
    except KeyError:
        error = None
    # Pull out the details if they're there
    if isinstance(error, dict):
        error = error.get('details')
    if error:
        click.echo('Error: ' + error)



import click
import helium
import util
import dpath.util as dpath

pass_service=click.make_pass_decorator(helium.Service)

@click.group()
def cli():
    """Operations on the user.
    """
    pass

def _tabulate(result):
    util.tabulate(result, [
        ('id', 'id'),
        ('name', 'attributes/name')
    ])


@cli.command()
@pass_service
def list(service):
    """Lists user information for the current user.

    Lists information associated withe the user for the current api key.
    """
    _tabulate([service.get_user().get('data')])


@cli.command()
@click.argument('user')
@click.password_option(confirmation_prompt=False)
@pass_service
def auth(service, user, password):
    """Authenticates the given user and password.

    Authenticates the given USER and password.
    If successful the API key for the account is printed
    """
    result = service.auth_user(user, password).get('data')
    click.echo(dpath.get(result,'meta/api-token'))



import sensor as _sensor
import helium
import util
import click
import dpath.util as dpath
import timeseries as ts

pass_service=click.make_pass_decorator(helium.Service)

@click.group()
def cli():
    """Operations on labels of sensors.
    """
    pass

def _tabulate(result):
    def _map_sensor_count(json):
        return len(dpath.get(json, 'relationships/sensor/data'))
    util.tabulate(result, [
        ('id', util.shorten_json_id),
        ('sensors', _map_sensor_count),
        ('name', 'attributes/name')
    ])

def _update_label_sensors(ctx, label, sensor, set_func):
    service = ctx.find_object(helium.Service)
    label = util.lookup_resource_id(service.get_labels, label)
    # Fetch the existing sensors
    sensors = service.get_label_sensors(label).get('data')
    sensor_ids = dpath.values(sensors, "*/id")
    # Look up full sensor ids for all given sensors
    sensor_list = [util.lookup_resource_id(service.get_sensors, sensor_id)
                   for sensor_id in sensor]
    # And perform the set operation and ensure we have a valid list
    sensor_ids = set_func(set(sensor_ids), set(sensor_list))
    if sensor_ids is None: sensor_ids = []
    service.update_label_sensors(label, sensor_ids)


@cli.command()
@click.argument('label', required=False)
@click.pass_context
def list(ctx, label):
    """Lists information on labels.

    Lists information on a given label or all labels in the organization
    """
    if label:
        ctx.invoke(_sensor.list, label=label)
    else:
        service = ctx.find_object(helium.Service)
        _tabulate(service.get_labels(include='sensor').get('data'))


@cli.command()
@click.argument('name', nargs=1)
@click.argument('sensor', nargs=-1)
@click.pass_context
def create(ctx, name, sensor):
    """Create a label.

    Creates a label with a given NAME and an (optional) list of SENSORs
    associated with that label.
    """
    service = ctx.find_object(helium.Service)
    label = service.create_label(name).get('data')
    label_id = label['id']
    if sensor:
        sensor_list = service.get_sensors().get('data')
        sensors = [util.lookup_resource_id(sensor_list, sensor_id)
                   for sensor_id in sensor]
        service.update_label_sensors(label_id, sensors)
    _tabulate([service.get_label(label_id, include='sensor').get('data')])


@cli.command()
@click.argument('label', nargs=-1)
@pass_service
def delete(service, label):
    """Delete a label.

    Deletes the LABEL with the given id
    """
    label_list = service.get_labels().get('data')
    label = [util.lookup_resource_id(label_list, label_id)
             for label_id in label]
    for entry in label:
        result = service.delete_label(entry)
        click.echo("Deleted: " +  entry if result.status_code == 204 else result)


@cli.command()
@click.argument('label',nargs=1)
@click.argument('sensor',nargs=-1)
@click.pass_context
def add(ctx, label, sensor):
    """Add sensors to a label.

    Adds a given list of SENSORs to the LABEL with the given id.
    """
    _update_label_sensors(ctx, label, sensor, set.union)
    ctx.invoke(_sensor.list, label=label)


@cli.command()
@click.argument('label', nargs=1)
@click.argument('sensor', nargs=-1)
@click.pass_context
def remove(ctx, label, sensor):
    """Remove sensors from a label.

    Removes a given list of SENSORs from the LABEL with the given id.
    """
    _update_label_sensors(ctx, label, sensor, set.difference)
    ctx.invoke(_sensor.list, label=label)


@cli.command()
@click.argument('label')
@ts.options()
@pass_service
def dump(service, label, format, **kwargs):
    """Dumps timeseries data to files.

    Dumps the timeseries data for all sensors in a given LABEL.

    One file is generated for each sensor with the sensor id as filename and the
    file extension based on the requested dump format
    """
    label = util.lookup_resource_id(service.get_labels, label)
    sensors = dpath.values(service.get_label_sensors(label), '/data/*/id')
    ts.dump(service, sensors, format, **kwargs)
